/*
Copyright 2012 Nathanael Hübbe
Contact: nathanael.huebbe@informatik.uni-hamburg.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _GNU_SOURCE

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <mpi.h>
#include <netcdf.h>
#ifdef NCIO_H_AVAILABLE
	#include <ncio.h>
#endif

const int kDimensionCount = 3;
const int kVariableCount = 40;
const int kRecordCount = 20;
const size_t kDimensionSizes[] = {1, 20000, 100};

// Use -r | --reduced to get reduced output (only the naked throughput in MiB/s is printed).
// Use -n | --no-fill to use NC_NOFILL
int main(int argc, char** argv) {
	int myProcId, procCount;
	int outputFile;
	int ret, intTrash;
	int varIds[kVariableCount], dimIds[kDimensionCount];
	size_t reducedDimensionSizes[kDimensionCount], startCoords[kDimensionCount];
	long i, floatCount;
	char name[100];	//I know this is bad, but I can't use glibc on the blizzard, so there's no asprintf available.
	struct timeval startTime, endTime;
	bool reducedOutput = false, useNoFill = false;
	
	MPI_Init(&argc, &argv);
	MPI_Comm_rank(MPI_COMM_WORLD, &myProcId);
	MPI_Comm_size(MPI_COMM_WORLD, &procCount);
	
	for(i = 1; i < argc; i++) {
		if(!strcmp(argv[i], "-r") || !strcmp(argv[i], "--reduced")) reducedOutput = true;
		else if(!strcmp(argv[i], "-n") || !strcmp(argv[i], "--no-fill")) useNoFill = true;
		else {
			if(!myProcId) {
				printf("Unknown argument: %s\nSupported options:\n", argv[i]);
				printf("\t-n | --no-fill Set NC_NOFILL for the file\n");
				printf("\t-r | --reduced Print only the naked throughput in MiB/s\n");
			}
			return -1;
		}
	}

	if(!reducedOutput && !myProcId) {
		#ifdef NCIO_H_AVAILABLE
			#ifdef DONT_USE_CACHE_FOR_OUTPUT
				printf("Not using cache for output.\n");
			#else
				printf("Using cache.\n");
			#endif
			#ifdef DONT_INITIALIZE_DATA
				printf("Not initializing data.\n");
			#else
				printf("Initializing data.\n");
			#endif
		#else
			printf("Can't figure out which (hacked) netcdf version we are using.\n");
		#endif
	}

	// Get the area that is to be written by this process.
	for(i = 0; i < kDimensionCount; i++) reducedDimensionSizes[i] = kDimensionSizes[i];
	reducedDimensionSizes[1] = ((myProcId + 1)*(uint64_t)kDimensionSizes[1])/procCount - (myProcId*(uint64_t)kDimensionSizes[1])/procCount;

	MPI_Barrier(MPI_COMM_WORLD);
	if(gettimeofday(&startTime, NULL)) return -1;

	// Open the files.
	if(snprintf(name, sizeof(name), "/work/ku0598/bench-%d.nc", myProcId) <= 0) return -3;
//	ret = nc__create(name, NC_CLASSIC_MODEL, 1024*1024, &chunkSize, &outputFile);
	#ifdef USE_NETCDF4
		ret = nc_create(name, NC_CLASSIC_MODEL | NC_NETCDF4, &outputFile);
	#else
		ret = nc_create(name, NC_CLASSIC_MODEL, &outputFile);
	#endif
	if(useNoFill) ret = nc_set_fill(outputFile, NC_NOFILL, &intTrash);

	//write the dimension descriptions
	floatCount = 1;
	for(i = 0; i < kDimensionCount; i++) {
		if(snprintf(name, sizeof(name), "dimension %li", i+1) <= 0) return -1;
		if(i) {
			if(nc_def_dim(outputFile, name, reducedDimensionSizes[i], &dimIds[i])) return -1;
			floatCount *= reducedDimensionSizes[i];
		} else {
			if(nc_def_dim(outputFile, name, NC_UNLIMITED, &dimIds[i])) return -1;
		}
	}

	//write the variable descriptions
	for(i = 0; i < kVariableCount; i++) {
		if(snprintf(name, sizeof(name), "variable %li", i+1) <= 0) return -1;
		if(nc_def_var(outputFile, name, NC_FLOAT, kDimensionCount, dimIds, &varIds[i])) return -1;
	}
	if(nc_enddef(outputFile)) return -1;

	//fake some data
	float* data = (float*)malloc(sizeof(float)*floatCount);
	for(i = 0; i < floatCount; i++) data[i] = (float)i;

	//write it many times
	for(i = 0; i < kDimensionCount; i++) startCoords[i] = 0;
	for(i = 0; i < kRecordCount; i++) {
		startCoords[0] = i;
		long curVar;
		for(curVar = 0; curVar < kVariableCount; curVar++) {
			nc_put_vara_float(outputFile, varIds[curVar], startCoords, reducedDimensionSizes, data);
		}
	}
	nc_close(outputFile);

	// Now the interesting part: How long did we take?
	MPI_Barrier(MPI_COMM_WORLD);
	if(!myProcId) {
		double totalFloatCount = 1, elapsedTime, throughput;
		if(gettimeofday(&endTime, NULL)) return -19;
		for(i = 1; i < kDimensionCount; i++) totalFloatCount *= kDimensionSizes[i];
		elapsedTime = (double)(endTime.tv_sec - startTime.tv_sec) + (endTime.tv_usec - startTime.tv_usec)/(double)1000000;
		throughput = totalFloatCount*kRecordCount*kVariableCount*4/(1024*1024*elapsedTime);
		if(reducedOutput) {
			printf("%.3f\n", throughput);
		} else {
			printf("Elapsed time: %.6fs\n", elapsedTime);
			printf("Speed: %.3fMiByte/s\n", throughput);
		}
	}
	
	MPI_Finalize();
	
	return 0;
}
